package aQute.service.library;

import java.io.*;
import java.net.*;
import java.util.*;

import aQute.struct.*;

public interface Library {
	String	VERSION		= "1.0.0";
	String	OSGI_GROUP	= "osgi";
	String	SHA_GROUP	= "sha";

	enum Phase {
		STAGING(false, false, false), LOCKED(true, false, false), MASTER(true, true, true), RETIRED(true, false, true), WITHDRAWN(
				true, false, true), UNKNOWN(true, false, false);

		boolean	locked;
		boolean	listable;
		boolean	permanent;

		private Phase(boolean locked, boolean listable, boolean permanent) {
			this.locked = locked;
			this.listable = listable;
			this.permanent = permanent;
		}

		public boolean isLocked() {
			return locked;
		}

		public boolean isListable() {
			return listable;
		}

		public boolean isPermanent() {
			return permanent;
		}
	}

	interface DownloadListener {
		void success(File file) throws Exception;

		void failed(File file, String reason) throws Exception;

		boolean progress(int percentage) throws Exception;
	}

	class Info extends struct {
		public String	version	= VERSION;	// TODO must be aligned with the
											// package version
		@Define(description = "Name of the library", pattern = "[\\w_][-\\w\\d_.]*")
		public String	name;
		public URI		location;
		public String	generation;
		public long		masterUpdates;
		public long		updates;
		public String	message;
		public boolean	approved;
	}

	class Relocation extends struct {
		public String	groupId;
		public String	artifactId;
		public String	version;
		public String	message;
	}

	class Revision extends SharedTypes.RepoPom {
		// SHA of resource
		public byte[]				_id;
		public URI					url;
		public byte[]				md5;
		public URI					pomUrl;
		public byte[]				pom;
		// Logical name. This is Bundle Symbolic Name
		// for OSGi bundles and groupid__artifactid for
		// maven artifacts that are not OSGi
		public String				title;
		public String				bsn;
		public String				baseline;
		public String				qualifier;
		public Phase				phase			= Phase.STAGING;

		public long					size;

		public String				receipt;										// Importer
																					// receipt
		public long					created;										// date
																					// of
																					// insertion
		public long					modified;										// when
																					// phase
																					// is
																					// changed
		public String				message;
		public String				owner;											// email
		public String				releaseSummary;

		public List<String>			signers;
		public Set<String>			category		= set();

		public Map<String,Object>	metadata		= new HashMap<String,Object>();
		public URI					icon;
		public List<String>			process			= list();
		public List<Requirement>	requirements	= list();
		public List<Capability>		capabilities	= list();
		public Set<String>			hashes			= set();
		public List<String>			errors			= list();
		public List<String>			warnings		= list();
		public List<String>			packages		= list();
		public List<URI>			repositories	= list();
		public String				mainClass;
		public String				docUrl;
		public Set<String>			keywords		= set();
		public Relocation			relocation;
	}

	class Namespace extends struct {
		public String				ns;
		public Map<String,Object>	ps	= new HashMap<String,Object>();
		public String				name;
	}

	class Requirement extends Namespace {}

	class Capability extends Namespace {}

	class RevisionRef extends struct {

		public RevisionRef() {}

		public RevisionRef(Revision revision) {
			this.revision = revision._id;
			this.bsn = revision.bsn;
			this.url = revision.url;
			this.md5 = revision.md5;
			this.phase = revision.phase;
			this.version = revision.version;
			this.classifier = revision.classifier;
			this.baseline = revision.baseline;
			this.qualifier = revision.qualifier;
			this.releaseSummary = revision.releaseSummary;
			if (revision.scm != null)
				this.tag = revision.scm.tag;
			this.name = revision.name;
			this.title = revision.title;
			this.description = revision.description;
			this.created = revision.created;
			this.errors = revision.errors.size();
			this.groupId = revision.groupId;
			this.artifactId = revision.artifactId;
			this.size = revision.size;
		}

		public byte[]	revision;
		public URI		url;
		public byte[]	md5;
		public String	groupId;
		public String	artifactId;
		public String	version;
		public String	classifier;
		public String	packaging;
		public String	bsn;
		public String	title;
		public String	name;
		public String	description;
		public String	baseline;
		public String	qualifier;
		public String	tag;
		public Phase	phase;
		public String	releaseSummary;
		public long		created;
		public int		errors;
		public long		size;
	}

	class Wiki extends struct {
		public String	text;
		public String	author;
		public long		modified;
	}

	class Program extends struct {
		// groupId:artifactId
		public byte[]				_id;
		public String				artifactId;
		public String				groupId;
		public String				bsn;
		public URI					icon;
		public Wiki					wiki;
		public int[]				rating		= new int[5];
		public long					modified;
		public List<RevisionRef>	revisions	= new ArrayList<Library.RevisionRef>();
		public Revision				last;
		public Set<String>			category	= set();
		public Set<String>			keywords	= set();
		public Set<String>			search		= set();

	}

	class Category extends struct {
		public String	_id;
		public String	summary;
		public Wiki		wiki;
		public URI		icon;
	}

	class Group extends Category {}

	Program getProgram(String groupId, String artifactId) throws Exception;

	// void master(RevisionRef rev);

	class Response extends struct {
		public Revision		revision;
		public List<String>	errors		= new ArrayList<String>();
		public List<String>	warnings	= new ArrayList<String>();
	}

	void update(Program program) throws Exception;

	void redo(String where, String process) throws Exception;

	boolean master(String bsn) throws Exception;

	Revision delete(String bsn);

	interface Find<T> extends Iterable<T> {
		Find<T> bsn(String bsn) throws Exception;

		Find<T> baseline(String version) throws Exception;

		Find<T> version(String version) throws Exception;

		Find<T> qualifier(String qualifier) throws Exception;

		Find<T> from(long date) throws Exception;

		Find<T> until(long date) throws Exception;

		Find<T> skip(int n) throws Exception;

		Find<T> limit(int n) throws Exception;

		Find<T> ascending(String field) throws Exception;

		Find<T> descending(String field) throws Exception;

		Find<T> where(String field, Object... args) throws Exception;

		Find<T> template(Revision rev) throws Exception;

		T one() throws Exception;

		T first() throws Exception;

		int count() throws Exception;

		Find<T> query(String query) throws Exception;

		Find<T> capability(String ns, String key, Object value) throws Exception;
	}

	Find< ? extends Program> findProgram() throws Exception;

	Find< ? extends Revision> findRevision() throws Exception;

	Info getInfo() throws Exception;

	Revision scan(URI uri) throws Exception;

	Revision stage(Revision revision) throws Exception;

	/**
	 * is stage( scan(url) )
	 * 
	 * @param url
	 * @return
	 * @throws Exception
	 */
	Revision stage(URI url) throws Exception;

	Revision rescan(String coordinates) throws Exception;

	Revision getRevision(byte[] sha) throws Exception;

	Revision getRevision(String groupId, String artifactId, String classifier, String version) throws Exception;

	Revision getRevision(String coordinates) throws Exception;

	Iterable< ? extends Revision> getRevisions(String coordinates) throws Exception;

	Category getCategory(String name) throws Exception;

	Iterable< ? extends Category> getCategories(String where) throws Exception;

	void update(Category c) throws Exception;

}
